#include "pzlgame.h"

using namespace game;

namespace pzlscript
{
	//Scripting for various map entities for the puzzel games :D
	//Eventually, when I figure out how I'm going to do map entites
	//That have proper physics, perhaps I'll add Newton game dynamics
	//Or ODE physics, Bullet is out of the question, too slow...
}
